// File for Belle2 slow control access with PCIe40 
#include "pcie40_b2slc.h"
#include "pcie40_ecs.h"
#include <unistd.h>
#include <stdio.h>
#include <vector>
#include <string.h>

// ===== 
// CALL C functions
// =====

int pcie40_writestream_internal( int dev , int ch , char * filename , int do_sleep ) ;

///////////////////////////////////////////////////////////////////////////
// Internal functions for slow control,, do not use outside of this library
///////////////////////////////////////////////////////////////////////////
int pcie40_writeFifoFillLevel( int dev , int ch ) {
  unsigned ret = 0 ;
  ret = ecs_read( dev , SLC_BAR , SLC_RFIFO_STATUS + ch * 0x20 ) ;
  return ( ret & 0x7F ) ; 
}

int pcie40_readFifoFillLevel( int dev , int ch ) {
  unsigned ret = ecs_read( dev , SLC_BAR , SLC_RFIFO_STATUS + ch * 0x20 ) ;
  ret = ( ret & 0xFF800 ) >> 11 ;
  return ret ;
}

int pcie40_resetWriteFifo( int dev , int ch ) {
  unsigned ret = 0 ;
  unsigned wle = 0 ; 
  // wait that the FIFO are empty
  int i ;
  for ( i=0 ; i<10 ; i++ ) {
    wle = pcie40_writeFifoFillLevel( dev , ch ) ; 
    if ( wle == 0 ) break;
    usleep( 10 ) ; //10 ms
  }

  if (i == 10) {
    printf( "Timeout reset write\n" ) ;
    // Do not return to perform the reset; return -1 ;
  }

  
  if ( ch < 24 ) {
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_0 , 0 ) ;
    if ( ret != 0 ) return -1 ;
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_0 , 
		     1 << ( SLC_WFIFO_RESET_BIT + ch ) ) ;
    if ( ret != 0 ) return -1 ;
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_0 , 0 ) ;
    if ( ret != 0 ) return -1 ;
  } else {
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_1 , 0 ) ;
    if ( ret != 0 ) return -1 ;
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_1 , 
		     1 << ( SLC_WFIFO_RESET_BIT + ch - 24 ) ) ;
    if ( ret != 0 ) return -1 ;
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_1 , 0 ) ;
    if ( ret != 0 ) return -1 ;
  }

  // Check that FIFO is really reset
  wle = pcie40_writeFifoFillLevel( dev , ch ) ; 
  wle = pcie40_writeFifoFillLevel( dev , ch ) ;

  if ( 0 != wle ) {
    printf( "Write FIFO not empty Level = %d\n" , wle ) ; 
    return -1 ; 
  }

  return ret ;
}

int pcie40_resetReadFifo( int dev , int ch ) {
  
  unsigned ret = 0 ;
  unsigned rle = 0 ; 
  // wait that the FIFO are empty
  int i ;
  for ( i=0 ; i<10 ; i++ ) {
    rle = pcie40_readFifoFillLevel( dev , ch ) ; 
    if ( rle == 0 ) break;
    usleep( 10 ) ; //10 ms
  }

  if (i == 10) {
    printf( "Timeout reset read\n" ) ;
  }

  if ( ch < 24 ) {
    ret = ecs_write( dev , SLC_BAR , SLC_RFIFO_RESET_ADD_0 , 0 ) ;
    if ( ret != 0 ) return -1 ;  
    ret = ecs_write( dev , SLC_BAR , SLC_RFIFO_RESET_ADD_0 , 
		     1 << ( SLC_RFIFO_RESET_BIT + ch ) ) ;
    if ( ret != 0 ) return -1 ;
    ret = ecs_write( dev , SLC_BAR , SLC_RFIFO_RESET_ADD_0 , 0 ) ;
    if ( ret != 0 ) return -1 ;
  } else {
    ret = ecs_write( dev , SLC_BAR , SLC_RFIFO_RESET_ADD_1 , 0 ) ;
    if ( ret != 0 ) return -1 ;  
    ret = ecs_write( dev , SLC_BAR , SLC_RFIFO_RESET_ADD_1 , 
		     1 << ( SLC_RFIFO_RESET_BIT + ch - 24 ) ) ;
    if ( ret != 0 ) return -1 ;
    ret = ecs_write( dev , SLC_BAR , SLC_RFIFO_RESET_ADD_1 , 0 ) ;
    if ( ret != 0 ) return -1 ;
  }

  // check that FIFO is empty 
  rle = pcie40_readFifoFillLevel( dev , ch ) ; 
  if ( 0 != rle ) { 
    printf( "Read FIFO not empty\n" ) ; 
    return -1 ; 
  }
  return ret ;
}

int pcie40_writeToFifo( int dev , int ch , std::vector< int > & data ) {
  std::vector< int >::iterator it ;
  unsigned ret = 0 ;
  unsigned fl = 0 ; 
  unsigned tries = 0 ; 
  while ( fl != data.size() ) {
    for ( it = data.begin() ; it != data.end() ; ++it ) {
      ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_ADD + ch * 0x20  , 
		       (int) ( (*it) & 0xFFFFFFFF ) ) ;
      if ( ret != 0 ) return -1 ;
    }

    // two reads are needed it seems, because the first one can be too close to the
    // previous write of the FIFO
    
    fl = pcie40_writeFifoFillLevel( dev , ch ) ; 
    if ( fl == data.size() ) return ret ;
    
    fl = pcie40_writeFifoFillLevel( dev , ch ) ; 
    if ( fl == data.size() ) return ret ; 
    
    // Reset Write FIFO 
    if ( ch < 24 ) {
      ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_0 , 0 ) ;
      if ( ret != 0 ) return -1 ;
      ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_0 , 
		       1 << ( SLC_WFIFO_RESET_BIT + ch ) ) ;
      if ( ret != 0 ) return -1 ;
      ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_0 , 0 ) ;
    if ( ret != 0 ) return -1 ;
    } else { 
      ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_1 , 0 ) ;
      if ( ret != 0 ) return -1 ;
      ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_1 , 
		       1 << ( SLC_WFIFO_RESET_BIT + ch - 24 ) ) ;
      if ( ret != 0 ) return -1 ;
      ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_RESET_ADD_1 , 0 ) ;
      if ( ret != 0 ) return -1 ;
    }

    tries++ ; 
    if ( tries == 100 ) { 
      printf( "Cannot write FIFO\n" ) ; 
      return -1 ;
    }
  }
  return ret ;
}

int pcie40_startEmit( int dev , int ch , int expected ) {
  int ret = 0 ;
  
  if ( expected != -1 ) { 
    // check that FIFO is written
    ret = pcie40_writeFifoFillLevel( dev , ch ) ;
    ret = pcie40_writeFifoFillLevel( dev , ch ) ;
    if ( expected != ret ) { 
      printf( "Write FIFO not written correctly. Number of words in FIFO: %d, number expected: %d\n" , ret , expected ) ;
      return -1 ; 
    }
  }

  if ( ch < 24 ) {
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD_0 , 0 ) ;
    if ( ret != 0 ) return -1 ;
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD_0 , 
		     1 << ( SLC_WFIFO_EMIT_BIT + ch ) ) ;
    if ( ret != 0 ) return -1 ;
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD_0 , 0 ) ;
    if ( ret != 0 ) return -1 ;
  } else { 
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD_1 , 0 ) ;
    if ( ret != 0 ) return -1 ;
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD_1 , 
		     1 << ( SLC_WFIFO_EMIT_BIT + ch - 24 ) ) ;
    if ( ret != 0 ) return -1 ;
    ret = ecs_write( dev , SLC_BAR , SLC_WFIFO_START_ADD_1 , 0 ) ;
    if ( ret != 0 ) return -1 ;
  }
  return ret ;
}

int pcie40_waitRead( int dev , int ch , unsigned int length ) {
  unsigned ret = 0 ;
  int i ;
  for ( i=0 ; i<100 ; i++ ) {
    ret = pcie40_readFifoFillLevel( dev , ch ) ;
    if ( ret == length ) break;
    usleep( 10 ) ; //10 us
  }

  if (i == 100) {
    printf( "Timeout, FIFO fill level = %d\n" , ret ) ;
    return -1 ;
  } 
  return 0 ; 
}

int pcie40_readData( int dev , int ch , std::vector< int > & result , int length ) {
  unsigned ret = 0 ;
  int i = 0 ;
 
  for ( i = 0 ; i < length ; ++i ) {
    ret =  ecs_read( dev , SLC_BAR , SLC_RFIFO_ADD + ch * 0x20 );
    result.push_back( ret ) ;
  }
  ret = pcie40_readFifoFillLevel( dev , ch ) ;
  if ( ret != 0 ) { printf( "Not all FIFO read\n" ) ; } 
  return 0 ;
}
/* ---------------------------------------------------------------------- *\
   readfee8
   returns -1 in case of error
\* ---------------------------------------------------------------------- */
int pcie40_readfee8( int dev , int ch , int adr) {
  // PCIe40
  if ( ( adr <=0 ) || ( adr > 0x7F ) ) return -1 ;
  // Reset the FIFO (Emit)
  unsigned ret = 0 ;
  ret = pcie40_resetWriteFifo( dev , ch ) ;
  if ( ret != 0 ) return ret ;

  // Fill the FIFO with the requested information: write MSB first
  std::vector< int > data ;
  data.push_back( 0xFFFE ) ; 
  data.push_back( 0x0001 ) ; 
  data.push_back( 0x8000 | ( adr & 0x7F ) ) ; 
  data.push_back( 0x000c ) ; 
  data.push_back( 0xEEEE ) ;

  ret = pcie40_writeToFifo( dev , ch , data ) ;
  if ( ret != 0 ) return ret ;

  // reset read FIFO
  ret = pcie40_resetReadFifo( dev , ch ) ;
  if ( ret != 0 ) return ret ;

  // start emit
  ret = pcie40_startEmit( dev , ch , 5 ) ;
  if ( ret != 0 ) return -1 ;

  // Wait for the result to come back (40 words in the fifo)
  ret = pcie40_waitRead( dev , ch , 5 ) ;
  if ( ret != 0 ) return -1 ;

  // Read the value 
  std::vector< int > result ; 
  ret = pcie40_readData( dev , ch , result , 5 ) ;

  int ret2 = result.at( 1 ) ; 
  int ret3 = result.at( 2 ) ; 
  // check address consistency
  if ( ( ( ret2 & 0x7F ) ) != adr ) return -1 ;

  return ( ret3 & 0xFF ) ;
}

//==============================================================================
// Write 8b
//==============================================================================
int pcie40_writefee8( int dev , int ch , int adr , int val ) {
  // PCIe40
  if ( ( adr <=0 ) || ( adr > 0x7F ) ) return -1 ;
  // Reset the FIFO
  unsigned ret = 0 ;
  ret = pcie40_resetWriteFifo( dev , ch ) ;
  if ( ret != 0 ) return ret ;

  // Fill the FIFO with the requested information: write MSB first
  std::vector< int > data ;
  data.push_back( 0xFFFF ) ; 
  data.push_back( 0x0001 ) ; 
  data.push_back( 0x8000 | ( adr & 0x7F ) ) ; 
  data.push_back( 0x0000 | ( val & 0xFF ) ) ;
  data.push_back( 0xEEEE ) ;
  ret = pcie40_writeToFifo( dev , ch , data ) ;
  if ( ret != 0 ) return ret ;

  // start emit
  ret = pcie40_startEmit( dev , ch , 5 ) ;
  if ( ret != 0 ) return -1 ;

  return 0 ;
}

//==============================================================================
// Read 32b
//==============================================================================
unsigned long pcie40_readfee32( int dev , int ch , int adr ) {
  // PCIe40
  if ( ( adr <0 ) || ( adr > 0xFFFF ) ) return -1 ;
  // Reset the FIFO
  unsigned ret = 0 ;
  ret = pcie40_resetWriteFifo( dev , ch ) ;
  if ( ret != 0 ) return ret ;
  
  // Fill the FIFO with the requested information: write MSB first
  std::vector< int > data ;
  data.push_back( 0xFFFB ) ; 
  data.push_back( 0x7000 | ( ( adr & 0xFF00 ) >> 8 ) ) ; 
  data.push_back( 0x7000 | ( adr & 0xFF ) ) ; 
  data.push_back( 0xEEEE ) ;

  ret = pcie40_writeToFifo( dev , ch , data ) ;
  if ( ret != 0 ) return ret ;

  // reset read FIFO
  ret = pcie40_resetReadFifo( dev , ch ) ;
  if ( ret != 0 ) return ret ;
  
  // start emit
  ret = pcie40_startEmit( dev , ch , 4 ) ;
  if ( ret != 0 ) return -1 ;
  
  // Wait for the result to come back
  ret = pcie40_waitRead( dev , ch , 7 ) ;
  if ( ret != 0 ) return -1 ;
  
  // Read the value 
  std::vector< int > result ;
  ret = pcie40_readData( dev , ch , result , 7 ) ;
  
  int t_ret1 = result.at(1);
  int t_ret2 = result.at(2);
  int t_ret3 = result.at(3);
  int t_ret4 = result.at(4);  
  int t_ret5 = result.at(5);
  int t_ret6 = result.at(6);

  if ( ( t_ret1 & 0x0000FFFF ) != 0xFFFF ) {
    printf( "Bad header\n" ) ;
    printf( "Data received: %X, %X, %X, %X, %X, %X\n" , t_ret1 , t_ret2 , t_ret3 , t_ret4 , t_ret5 , t_ret6 ) ; 
    return -1 ; 
  }
  if ( ( t_ret6 & 0x0000FFFF ) != 0xEEEE ) { 
    printf( "Bad trailer\n" ) ;
    printf( "Data received: %X, %X, %X, %X, %X, %X\n" , t_ret1 , t_ret2 , t_ret3 , t_ret4 , t_ret5 , t_ret6 ) ; 
    return -1 ;
  }

  int add1 = ( t_ret2 & 0xFF00 ) >> 8 ;
  int add2 = ( t_ret3 & 0xFF00 ) >> 8 ;
  int add3 = ( t_ret4 & 0xFF00 ) >> 8 ;
  int add4 = ( t_ret5 & 0xFF00 ) >> 8 ;
  
  if ( ( add1 & 0xF8 ) != 0x10 ) { 
    printf( "Bad address1 %X\n" , add1 ) ; 
    return -1 ;
  }
  if ( ( add2 & 0xF8 ) != 0x10 ) { 
    printf( "Bad address2 %X\n" , add2 ) ; 
    return -1 ;
  }
  if ( ( add3 & 0xF8 ) != 0x10 ) {
    printf( "Bad address3 %X\n" , add3 ) ; 
    return -1 ;
  }
  if ( ( add4 & 0xF8 ) != 0x10 ) {
    printf( "Bad address4 %X\n" , add4 ) ; 
    return -1 ;
  }

  add1 += 0x58 ;
  add2 += 0x58 ;
  add3 += 0x58 ;
  add4 += 0x58 ;
  
  if ( ( add2 != 0x6C ) || ( add1 != 0x6D ) || ( add4 != 0x6A ) || ( add3 != 0x6B ) )
    {
      printf( "Unexpected address %X %X %X %X\n" , add1 , add2 , add3 , add4 ) ;
      return -1 ;
    }

  int res1 = ( t_ret2 & 0xFF ) ;
  int res2 = ( t_ret3 & 0xFF ) ;
  int res3 = ( t_ret4 & 0xFF ) ;
  int res4 = ( t_ret5 & 0xFF ) ;

  return 0xFFFFFFFF & ( ( res1 << 24 ) | ( res2 << 16 ) | ( res3 << 8 ) | ( res4 ) ) ;
}

//==============================================================================
// Write 32b
//==============================================================================
int pcie40_writefee32( int dev , int ch , int adr , int val ) { 
  // PCIe40
  if ( ( adr <0 ) || ( adr > 0xFFFF ) ) return -1 ;
  // Reset the FIFO
  unsigned ret = 0 ;
  ret = pcie40_resetWriteFifo( dev , ch ) ;
  if ( ret != 0 ) return ret ;

  int val1 = ( val & 0xFF ) ;
  int val2 = ( val & 0xFF00 ) >> 8 ;
  int val3 = ( val & 0xFF0000 ) >> 16 ;
  int val4 = ( val & 0xFF000000 ) >> 24 ;

  // Fill the FIFO with the requested information: write MSB first
  std::vector< int > data ;
  data.push_back( 0xFFFC ) ; 
  data.push_back( 0x7000 | ( ( adr & 0xFF00 ) >> 8 ) ) ; 
  data.push_back( 0x7000 | ( adr & 0xFF ) ) ; 
  data.push_back( 0x7000 | val4 ) ; 
  data.push_back( 0x7000 | val3 ) ; 
  data.push_back( 0x7000 | val2 ) ; 
  data.push_back( 0x7000 | val1 ) ;
  data.push_back( 0xEEEE ) ;

  ret = pcie40_writeToFifo( dev , ch , data ) ;
  if ( ret != 0 ) return ret ;

  // start emit
  ret = pcie40_startEmit( dev , ch , 8 ) ;
  if ( ret != 0 ) return -1 ;

  return ret;
}


int pcie40_writestream_klm( int dev , int ch , char * filename ) { 
  return pcie40_writestream_internal( dev , ch , filename , 0 ) ;
}

int pcie40_writestream_arich( int dev , int ch , char * filename ) { 
  return pcie40_writestream_internal( dev , ch , filename , 1 ) ;
}

int pcie40_writestream( int dev , int ch , char *filename , const char * detector )  { 
  if ( strcmp( detector , "KLM" ) == 0 ) { 
    return pcie40_writestream_klm( dev , ch , filename ) ;
  } else if ( strcmp( detector , "ARICH" ) == 0 ) { 
    return pcie40_writestream_arich( dev , ch , filename ) ; 
  } 
  printf( "ERROR : detector not supported\n" ) ;
  return -1 ; 
} 

int pcie40_writestream_internal( int dev , int ch , char * filename , int do_sleep ) {
  // PCIe40
  // Check if the file exists
  const char * mode = "r" ; 
  FILE * fp = fopen( filename , mode ) ;
  if ( 0 == fp ) {
    printf( "The file does not exist\n" ) ;
    return -1 ; 
  }

  unsigned ret = 0 ;
 
  // Read the file
  int c = 0 ;
  int count = 0 ;
  int count_sleep = 0 ;
  std::vector< int > data ; 
  while ((c = getc(fp)) != EOF) {
    if ( count == 0 ) // Stream write : insert FFFD
      data.push_back( 0xFFFD ) ; 
    count++ ; count_sleep++ ;
    data.push_back( (int) ( ( ( 0x70 << 8 ) | ( c & 0xFF ) ) & 0xFFFFFFFF ) ) ;

    if ( 6 == count ) {
      data.push_back( 0xEEEE ) ;
      // Reset the FIFO
      ret = pcie40_resetWriteFifo( dev , ch ) ;
      if ( ret != 0 ) { 
	fclose( fp ) ; 
	return -1 ;
      }

      ret = pcie40_writeToFifo( dev , ch , data ) ;
      if ( ret != 0 ) { 
        fclose( fp ) ; 
        return -1 ;
      }    
      ret = pcie40_startEmit( dev , ch , -1 ) ;
      if ( ret != 0 ) { 
        fclose( fp ) ; 
        return -1 ;
      }  
      if ( ( count_sleep >= 100 ) && ( do_sleep ) ) {
	usleep( 1 ) ; 
	count_sleep = 0 ;
      }
      count = 0 ; 
      data.clear() ;
    }
  }
  fclose( fp ) ;

  // End of the file -> 0xEEEE 
  data.push_back( 0xEEEE ) ; 

  // Reset the FIFO
  ret = pcie40_resetWriteFifo( dev , ch ) ;
  
  if ( ret != 0 ) { 
    fclose( fp ) ; 
    return -1 ;
  }
  
  ret = pcie40_writeToFifo( dev , ch , data ) ;
  if ( ret != 0 ) return -1 ;

  // start emit
  ret = pcie40_startEmit( dev , ch , -1 ) ; 
  if ( ret != 0 ) return -1 ;

  return ret ;
}

int pcie40_writebytestream( int dev, int ch, int len, const char *bytes  )
{
  unsigned ret = 0 ;
 
  // Read the file
  int c = 0 ;
  int i = 0 ;
  int count = 0 ;
  std::vector< int > data ; 
  while ( len != i ) {
    if ( count == 0 ) // Stream write : insert FFFD
      data.push_back( 0xFFFD ) ; 
    count++ ; 
    c = bytes[ i ] ; 
    i++ ; 
    data.push_back( (int) ( ( ( 0x70 << 8 ) | ( c & 0xFF ) ) & 0xFFFFFFFF ) ) ;

    if ( 6 == count ) {
      data.push_back( 0xEEEE ) ;
      // Reset the FIFO
      ret = pcie40_resetWriteFifo( dev , ch ) ;
      
      if ( ret != 0 ) return -1 ;

      ret = pcie40_writeToFifo( dev , ch , data ) ;
      if ( ret != 0 ) return -1 ;
      ret = pcie40_startEmit( dev , ch , -1 ) ;
      if ( ret != 0 ) return -1 ;

      count = 0 ; 
      data.clear() ;
    }
  }

  // End of the file -> 0xEEEE 
  data.push_back( 0xEEEE ) ; 

  // Reset the FIFO
  ret = pcie40_resetWriteFifo( dev , ch ) ;
  
  if ( ret != 0 ) return -1 ;
  
  ret = pcie40_writeToFifo( dev , ch , data ) ;
  if ( ret != 0 ) return -1 ;

  // start emit
  ret = pcie40_startEmit( dev , ch , -1 ) ; 
  if ( ret != 0 ) return -1 ;

  return ret ;
}
